/*
	(c) 2021 Dmitry Grinberg   https://dmitry.gr
	Non-commercial use only OR licensing@dmitry.gr
*/

#include "samda1e16b.h"
#include <string.h>
#include "printf.h"
#include "ufat.h"
#include "sd.h"

#define VERBOSE						0

#define FLASH_PAGES_PER_ROW			4			//this means that app must start on 256 byte boundary

extern uint32_t __bl_start[], __app_start[];
static uint8_t __attribute__((aligned(4))) mBuf[512];

void prPutchar(char chr)
{
	volatile uint32_t *addr = (volatile uint32_t*)ZWT_ADDR;
			
	while (addr[0] & 0x80000000ul);
	addr[0] = 0x80000000ul | (uint8_t)chr;
}

static void led(bool on)	//also SD nCS so do not do it while SD is being used
{
	if (on)
		PORT->Group[0].OUTCLR.reg = PORT_PA28;
	else
		PORT->Group[0].OUTSET.reg = PORT_PA28;
}

static void delay(uint32_t usec)		//at 8MHz for C-M0+
{
	asm volatile(
		".syntax unified	\n\t"
		"1:					\n\t"
		"	subs %0, #1		\n\t"		//1 cy
		"	bmi	2f			\n\t"		//1 cy when not taken
		"	nop				\n\t"		//1 cy
		"	nop				\n\t"		//1 cy
		"	nop				\n\t"		//1 cy
		"	nop				\n\t"		//1 cy
		"	b   1b			\n\t"		//2 cy when taken
		"2:					\n\t"
		:"+l"(usec)
		:
		:"cc"
	);
}

static void hwInit(void)
{
	#define SET_MUX(_idx, _val)		do { if ((_idx) & 1) PORT->Group[0].PMUX[(_idx) / 2].bit.PMUXO = (_val); else PORT->Group[0].PMUX[(_idx) / 2].bit.PMUXE = (_val); } while (0)
	
	GCLK->CLKCTRL.reg = GCLK_CLKCTRL_ID_SERCOM3_CORE |
						GCLK_CLKCTRL_GEN_GCLK0 |
						GCLK_CLKCTRL_CLKEN;
	PM->APBCMASK.reg |= PM_APBCMASK_SERCOM3;
	
	PORT->Group[0].DIRSET.reg =	PORT_PA14 |					//nCS for ram
								PORT_PA28 |					//nCS for SD
								PORT_PA22 | PORT_PA23;	 	//SERCOM3
	PORT->Group[0].DIRCLR.reg = PORT_PA19;					//SERCOM3

	//function outputs
	PORT->Group[0].PINCFG[PIN_PA22].reg = PORT_PINCFG_INEN | PORT_PINCFG_DRVSTR | PORT_PINCFG_PMUXEN;
	PORT->Group[0].PINCFG[PIN_PA23].reg = PORT_PINCFG_INEN | PORT_PINCFG_DRVSTR | PORT_PINCFG_PMUXEN;

	//function inputs
	PORT->Group[0].PINCFG[PIN_PA19].reg = PORT_PINCFG_INEN | PORT_PINCFG_PMUXEN | PORT_PINCFG_PULLEN;
	
	//gpio outputs
	PORT->Group[0].PINCFG[PIN_PA14].reg = PORT_PINCFG_INEN | PORT_PINCFG_DRVSTR;
	PORT->Group[0].PINCFG[PIN_PA28].reg = PORT_PINCFG_INEN | PORT_PINCFG_DRVSTR;
	
	//pull and output values
	PORT->Group[0].OUTCLR.reg = PORT_PA19;
	PORT->Group[0].OUTSET.reg = PORT_PA14 | PORT_PA28;
	
	//functions
	SET_MUX(PIN_PA19, 3);
	SET_MUX(PIN_PA22, 2);
	SET_MUX(PIN_PA23, 2);
	
	#undef SET_MUX
}

bool ufatExtRead(uint32_t sector, uint16_t offset, uint8_t len, uint8_t* buf)
{
	static uint32_t mSec = 0xffffffff;
	const volatile uint8_t *from;
	
	if (sector != mSec && !sdSecRead(sector, mBuf))
		return false;
	mSec = sector;
	
	from = mBuf + offset;
	while (len--)
		*buf++ = *from++;
	
	return true;
}

static bool findUpdateFile(uint32_t *sizeP, uint32_t maxSize)
{
	uint16_t n = 0, id;
	uint8_t flags;
	char name[12];
	uint32_t sz;
	
	for (n = 0; ufatGetNthFile(n, name, &sz, &flags, &id); n++) {
		if (VERBOSE)
			pr("'%s' %08xh bytes %02xh flags @ 0x%04xh\n", name, sz, flags, id);
		
		if (!memcmp(name, "FIRMWAREBIN", sizeof(name)) && sz > 1024 && sz < maxSize) {
			
			if (sizeP)
				*sizeP = sz;
			
			return ufatOpen(id);
		}
	}
	
	return false;
}

static bool flashPrvErzRow(uint32_t addr)
{
	NVMCTRL->INTFLAG.reg = NVMCTRL_INTFLAG_READY | NVMCTRL_INTFLAG_ERROR;
	
	NVMCTRL->ADDR.reg = addr / 2 /* 16 bit word address */;
	NVMCTRL->CTRLA.reg = NVMCTRL_CTRLA_CMDEX_KEY | NVMCTRL_CTRLA_CMD_ER;
	while (!NVMCTRL->INTFLAG.bit.READY);
	
	return !NVMCTRL->INTFLAG.bit.ERROR;
}

static bool flashPrvWritePage(uint32_t addr, const uint8_t *srcP, uint32_t pageSize)	//assumes src is 32-bit aligned
{
	volatile uint32_t *dst = (volatile uint32_t*)addr;
	const uint32_t *src = (const uint32_t*)srcP;
	uint32_t i;
	
	NVMCTRL->INTFLAG.reg = NVMCTRL_INTFLAG_READY | NVMCTRL_INTFLAG_ERROR;
	
	for (i = 0; i < pageSize / sizeof(uint32_t); i++)
		*dst++ = *src++;
	
	while (!NVMCTRL->INTFLAG.bit.READY);
	
	return !NVMCTRL->INTFLAG.bit.ERROR;
}

static bool performUpdate(uint32_t flashAddr, uint32_t numPages, uint32_t pageSize)
{
	uint32_t pgIdx, inSecOfst = sizeof(mBuf), secIdx = 0, numSec = 0, flashAddrStart = flashAddr;
	uint8_t __attribute__((aligned(4))) firstPage[pageSize];
	uint8_t *src;
	
	for (pgIdx = 0; pgIdx < numPages; pgIdx++, flashAddr += pageSize, inSecOfst += pageSize, src += pageSize) {
		
		if (inSecOfst == sizeof(mBuf)) {		//current sector buffer is used up - we need to read the next sec
			
			if (!numSec) {						//current sector range is used up - we need a new one
				
				if (!ufatGetNextSectorRange(&secIdx, &numSec)) {
					
					if (VERBOSE)
						pr("cannot get next sector range\n");
					return false;
				}
			}
			
			if (VERBOSE)
				pr("  for page %xh reading sec 0x%08x\n", pgIdx, secIdx);
			if (!sdSecRead(secIdx, mBuf)) {
							
				if(VERBOSE)
					pr("cannot read sector\n");
				return false;
			}
			secIdx++;
			numSec--;
			inSecOfst = 0;
			src = mBuf;
		}
		
		if (!(pgIdx % FLASH_PAGES_PER_ROW)) {
			if (VERBOSE)
				pr(" erz 0x%08x\n", flashAddr);
			if (!flashPrvErzRow(flashAddr)) {
				
				if (VERBOSE)
					pr("Failed to erase page\n");
				return false;
			}
		}
		
		if (VERBOSE)
			pr(" wri 0x%08x: %02x %02x %02x %02x\n", flashAddr, src[0], src[1], src[2], src[3]);
		if (!pgIdx) {	//first page preserved for later
			
			memcpy(firstPage, src, pageSize);
		}
		else if (!flashPrvWritePage(flashAddr, src, pageSize)) {
			
			if (VERBOSE)
				pr("Failed to write page\n");
			return false;
		}
	}
	
	if (VERBOSE)
		pr("doing last 0x%08x -> 0x%08x + 0x%08x\n", firstPage, flashAddrStart, pageSize);
	//if we got this far, we're done so far - write last page
	return flashPrvWritePage(flashAddrStart, firstPage, pageSize);
}

static void victoryDance(void)
{
	uint32_t freqMax = 500000, freqMin = 10000, freq = freqMax, nTimes = 2, i;
	bool down = true;
	
	while (1) {

		while (1) {
	
			for (i = 0; i < nTimes; i++) {
				led(true);
				delay(freq);
				led(false);
				delay(freq);
			}
			if (down) {
				
				freq /= 2;
				nTimes *= 2;
				if (freq < freqMin)
					break;
			}
			else {
				
				freq *= 2;
				nTimes /= 2;
				if (freq >= freqMax)
					break;
			}
		}
		down = !down;
	}
}

static void showError(uint32_t errNo)
{
	uint32_t i;
	
	while (1) {
		
		for (i = 0; i < errNo; i++) {
			
			led(true);
			delay(500000);
			led(false);
			delay(500000);
		}
		
		delay(2000000);
	}
}

static uint32_t tryPerformUpdate(void)	//return err no on error, does not return on success
{
	uint32_t pageShift = 3 + NVMCTRL->PARAM.bit.PSZ, pageSize = 1 << pageShift, pageMask = pageSize - 1, numPages = NVMCTRL->PARAM.bit.NVMP;
	uint32_t appBase = (uint32_t)__app_start, blBase = (uint32_t)__bl_start, appSpace = (numPages << pageShift) - (appBase - blBase);
	uint32_t updateFileSz;
		
	if (!sdCardInit()) {
		
		if (VERBOSE)
			pr("SD init fail\n");
		return 1;
	}
	
	if (!ufatMount()) {
		
		if (VERBOSE)
			pr("mount fail\n");
		return 2;
	}
	
	if (!findUpdateFile(&updateFileSz, appSpace)) {
		
		if (VERBOSE)
			pr("cannot open update file\n");
		return 3;
	}
	
	if (!performUpdate(appBase, (updateFileSz + pageSize) >> pageShift, pageSize)) {
		
		if (VERBOSE)
			pr("cannot perofrm update\n");
		return 4;
	}
	
	if (VERBOSE)
		pr("update performed\n");
	victoryDance();
	
	return 0;
}

static bool isAppValid(void)
{
	uint32_t flashSz = NVMCTRL->PARAM.bit.NVMP << (3 + NVMCTRL->PARAM.bit.PSZ);
	uint32_t allegedSP = __app_start[0], allegedPC = __app_start[1];
	
	if (allegedSP % 4)											//SP should be word aligned;
		return false;
	if (!(allegedPC % 2))										//PC must be odd
		return false;
	if (allegedSP < 0x20000000 || allegedSP >= 0x20002000)		//SP must be in RAM
		return false;
	if (allegedPC < 8 || allegedPC >= flashSz)					//PC must be in flash
		return false;
	
	return true;
}

static void doBoot(void)
{
	SCB->VTOR = (uint32_t)__app_start;
	asm volatile(
		"	ldmia %0, {r0, r1}	\n\t"
		"	mov   sp, r0		\n\t"
		"	bx    r1			\n\t"
		::"l"(__app_start)
	);
}

void __attribute__((noreturn, used)) micromain(void)
{
	volatile uint32_t t = 0;
	uint32_t errNo;
	
	SYSCTRL->OSC8M.bit.PRESC = 0;		//goto 8M
	for (t = 0; t < 0x8000; t += 2)
		t--;
	hwInit();
	if (VERBOSE)
		pr("go\n");
	
	errNo = tryPerformUpdate();
	if (isAppValid())
		doBoot();
	else
		showError(errNo);
	
	while (1);
}

void __attribute__((used)) report_hard_fault(uint32_t* regs, uint32_t ret_lr, uint32_t *user_sp)
{
	uint32_t *push = (ret_lr == 0xFFFFFFFD) ? user_sp : (regs + 8), *sp = push + 8;
	unsigned i;
	
	prRaw("== HARDFAULT ==\n");
	for (i = 0; i < 4; i++)
		prRaw("R%02xh=0x%08x\n", i, (unsigned)*push++);
	for (i = 0; i < 8; i++)
		prRaw("R%02xh=0x%08x\n", i + 4, (unsigned)*regs++);
	prRaw("R%02xh=0x%08x\n", 12, (unsigned)*push++);
	prRaw("R%02xh=0x%08x\n", 13, (unsigned)sp);
	for (i = 0; i < 3; i++)
		prRaw("R%02xh=0x%08x\n", i + 14, (unsigned)*push++);		//R16 is SR
	prRaw("== HARDFAULT ==\n");
	while (1);
}

void __attribute__((noreturn, naked, noinline)) HardFault_Handler(void)
{
	asm volatile(
			"mov  r0, r8				\n\t"
			"mov  r1, r9				\n\t"
			"mov  r2, r10				\n\t"
			"mov  r3, r11				\n\t"
			"push {r0-r3}				\n\t"
			"push {r4-r7}				\n\t"
			"mov  r0, sp				\n\t"
			"mov  r1, lr				\n\t"
			"mrs  r2, PSP				\n\t"
			"bl   report_hard_fault		\n\t"
			:::"memory");
}
