/*
	(c) 2021 Dmitry Grinberg   https://dmitry.gr
	Non-commercial use only OR licensing@dmitry.gr
*/

#include <sys/stat.h>
#include <string.h>
#include <stdio.h>
#include "printf.h"
#include "vfs.h"

#define MAX_VOLUMES					1
#define MAX_FILES					2

#define SUPERBLOCK_SECTOR			16
#define MAX_SUPERBLOCK_SECTORS		16
#define UFS_MAGIC					0x00011954

#define ROOT_DIR_INODE				2	//for historical reasons



#define DIRECT_BLKS_PER_INODE		12
#define INDIR_BLKS_PER_INODE		3	




//all aligned and all LE
struct CylSummary {
	uint32_t numDirs;
	uint32_t numFreeBlocks;
	uint32_t numFreeInodes;
	uint32_t numFreeFrags;
};



struct UfsSuper0 {
	uint32_t rfu[2];				//0x000
	uint32_t super0ofst;			//0x008	address of this block in sectors
	uint32_t cylOfst;				//0x00c address of the cylinder block in sectors
	uint32_t inodesOfst;			//0x010 address of inodes in sectors
	uint32_t firstDataOfst;			//0x014 address of first actual data (after cylinder group info) in sectors
	uint32_t cgOffset;				//0x018
	uint32_t cgMask;				//0x01c
	uint32_t lastWrittenTime;		//0x020
	uint32_t numBlocks;				//0x024	in the entire fs
	uint32_t numDataBlocks;			//0x028	in the entire fs
	uint32_t numCylGroups;			//0x02c
	
	uint32_t basicBlockSz;			//0x030
	uint32_t fragBlockSz;			//0x034
	uint32_t fragCt;				//0x038	frags per frag block
	
	uint32_t misc0[3];				//0x03c things we can live without
	
	uint32_t bMask;					//0x048 convenient thigns mkfs left for us
	uint32_t fMask;					//0x04c convenient thigns mkfs left for us
	uint32_t bFhift;				//0x050 convenient thigns mkfs left for us
	uint32_t fShift;				//0x054 convenient thigns mkfs left for us
	
	uint32_t misc1[2];				//0x058 things we can live without
	
	uint32_t fragShift;				//0x060 convenient thigns mkfs left for us
	uint32_t fsToDb;				//0x064 convenient thigns mkfs left for us
	
	uint32_t misc2[12];				//0x068 things we can live without
	
	uint32_t cylGrpSummaryOfst;		//0x098 in sectors
	uint32_t cylGrpSummarySz;		//0x09c
	
	uint32_t cylGroupSz;			//0x0a0
	uint32_t trackPerCyl;			//0x0a4
	uint32_t secsPerTrack;			//0x0a8
	uint32_t secsPerCy;				//0x0ac
	uint32_t numCyl;				//0x0b0 in the fs

	uint32_t cylsPerGroup;			//0x0b4
	uint32_t inodesPerGroup;		//0x0b8
	uint32_t blksPerGroup_M_fsFrag;	//0x0bc	M is mul
	
	struct CylSummary allSummary;	//0x0c0
	uint8_t flags[4];				//0x0d0
	
	char lastMountedAt[300];		//0x0d4
};	//0x200


struct UfsSuper1 {
	char lastMountedAtPt2[200];		//0x000
	uint8_t misc1[4];				//0x0c8
	uint32_t timeLastCheck;			//0x0cc
	uint32_t uid;					//0x0d0
	uint32_t misc2[33];				//0x0d4
	uint32_t cpc;					//0x158
	uint16_t posTblRawStart[82];	//0x15c  really is [32][8]
};

struct UfsSuper2 {
	uint16_t posTblRawStart[174];	//0x000
	uint32_t magic;					//0x15c
	uint8_t blkListp[];				//0x160
};

struct UfsTimeval {
	uint32_t seconds;				//0x00
	uint32_t micros;				//0x08
};


struct UfsInode {
	uint16_t mode	:12;							//0x00
	uint16_t type	: 4;
	uint16_t numLinks;								//0x02 number of hardlinks
	uint16_t ownerUid;								//0x04
	uint16_t ownerGid;								//0x06
	uint64_t size;									//0x08
	struct UfsTimeval access;						//0x10
	struct UfsTimeval mod;							//0x18
	struct UfsTimeval create;						//0x20
	uint32_t directBlocks[DIRECT_BLKS_PER_INODE];	//0x28
	uint32_t indirectBlocks[INDIR_BLKS_PER_INODE];	//0x58
	uint32_t flags;									//0x64
	uint32_t numBlocks;								//0x68
	uint32_t revNo;									//0x6c
	uint32_t rfu[4];								//0x70
};

struct UfsDirEntry {								//on nay 4 byte boundary
	uint32_t inode;									//0x00 set to 0 if this entry is empty
	uint16_t structSz;								//0x04 how big this struct is in bytes, must be multiple of 4, can be quite oversized
	uint16_t nameLen;								//0x06 how long the name is, not including zero term, but zero-term itself is guaranteed
	char name[];									//up to UFS_MAX_NAME_LEN + 1
};

struct FfsDirEntry {								//on nay 4 byte boundary
	uint32_t inode;									//0x00 set to 0 if this entry is empty
	uint16_t structSz;								//0x04 how big this struct is in bytes, must be multiple of 4, can be quite oversized
	uint8_t type;									//0x06
	uint8_t nameLen;								//0x07 how long the name is, not including zero term, but zero-term itself is guaranteed
	char name[];									//up to FFS_MAX_NAME_LEN + 1
};

struct VfsFile {
	struct VfsVolume *vol;
	uint32_t directBlocks[DIRECT_BLKS_PER_INODE];
	uint32_t indirectBlocks[INDIR_BLKS_PER_INODE];
	uint32_t size;
	
	uint32_t curPos;
	uint32_t curBlk;
	uint16_t curIndirIndirIdx;
	uint8_t curDirIdx;		//must be DIRECT_BLKS_PER_INODE to indicate we're indirect
	uint8_t curIndirIdx	: 2;
	uint8_t fileType	: 6;
};

struct VfsVolume {
	VfsReadSecProc readF;
	void *readD;
	
	//data
	uint32_t blksPerGroup_M_fsFrag, cgMask, cgOffset, inodesOfst, firstDataOfst, numCylGroups, inodesPerGroup, basicBlockSz;
	
	uint8_t fsToDb;
};

struct VfsFileOpenInfo {
	uint32_t inode;
};


static uint8_t *mBuf;
struct VfsVolume *mBufVol;
static uint32_t mBufSec;

struct VfsVolume mVolumes[MAX_VOLUMES];
struct VfsFile mFiles[MAX_FILES];
static uint8_t mVolumesAvail = (1 << MAX_VOLUMES) - 1;
static uint8_t mFilesAvail = (1 << MAX_FILES) - 1;


static uint8_t *mBuf;
struct VfsVolume *mBufVol;
static uint32_t mBufSec;


static bool ufsPrvReadSectorEx(struct VfsVolume *vol, uint32_t sec, void* dst)
{
	if (dst == mBuf)	//should not happen
		mBufVol = NULL;
	
	return vol->readF(vol->readD, sec, dst);
}

static void* ufsPrvReadSector(struct VfsVolume *vol, uint32_t sec)
{
	if (mBufVol == vol && mBufSec == sec)
		return mBuf;
	
	mBufVol = NULL;
	if (!ufsPrvReadSectorEx(vol, sec, mBuf))
		return NULL;
	
	mBufVol = vol;
	mBufSec = sec;
	
	return mBuf;
}

static uint32_t ufsPrvGetCylGrpStart(struct VfsVolume *vol, uint32_t cylGroupIdx)		//in blocks
{
	return vol->blksPerGroup_M_fsFrag * cylGroupIdx + vol->cgOffset * (cylGroupIdx &~ vol->cgMask);
}

static uint32_t ufsPrvGetInodesStart(struct VfsVolume *vol, uint32_t cylGroupIdx)		//in blocks
{
	return ufsPrvGetCylGrpStart(vol, cylGroupIdx) + vol->inodesOfst;
}

static uint32_t ufsPrvGetDataStart(struct VfsVolume *vol, uint32_t cylGroupIdx)			//in blocks
{
	return ufsPrvGetCylGrpStart(vol, cylGroupIdx);		//yes, without  + vol->firstDataOfst
}

static uint32_t ufsPrvBlockToSector(struct VfsVolume *vol, uint32_t block)
{
	return block << vol->fsToDb;
}

static uint32_t ufsPrvDataBlockNoToSector(struct VfsVolume *vol, uint32_t dataBlockNo)
{
	uint32_t dataBlocksCg, dataBlockIndexInCg;
	
	//this is a TODO since i dont seem to know how to find num data blocks per CG...
	dataBlocksCg = 0;
	dataBlockIndexInCg = dataBlockNo;
	
	return ufsPrvBlockToSector(vol, ufsPrvGetDataStart(vol, dataBlocksCg) + dataBlockIndexInCg);
}

static void ufsPrvDumpInode(struct VfsVolume *vol, const struct UfsInode *in)
{
	static const char *names[] = {
		[VFS_TYPE_NONE] = "none",
		[VFS_TYPE_PORT] = "port",
		[VFS_TYPE_CDEV] = "cdev",
		[VFS_TYPE_DIR] = "dir ",
		[VFS_TYPE_BDEV] = "bdev",
		[VFS_TYPE_FILE] = "file",
		[VFS_TYPE_SYMLINK] = "sym ",
		[VFS_TYPE_SOCKET] = "sock",
		[VFS_TYPE_PIPE] = "pipe",
	};
	const char *typ = names[in->type];
	
	if (!typ)
		typ = "XXXX";
	
	pr("  inode: %s(%u) %c%c%c%c%c%c%c%c%c%c%c (%u links) size %u, %u blocks, first 0x%x, which is ofst 0x%08x, second 0x%x, which is ofst 0x%08x\n", typ, in->type,
		(in->mode & S_IRUSR) ? 'r' : '-',
		(in->mode & S_IWUSR) ? 'w' : '-',
		(in->mode & S_IXUSR) ? 'x' : '-',
		(in->mode & S_IRGRP) ? 'r' : '-',
		(in->mode & S_IWGRP) ? 'w' : '-',
		(in->mode & S_IXGRP) ? 'x' : '-',
		(in->mode & S_IROTH) ? 'r' : '-',
		(in->mode & S_IWOTH) ? 'w' : '-',
		(in->mode & S_IXOTH) ? 'x' : '-',
		(in->mode & S_ISUID) ? 's' : '-',
		(in->mode & S_ISGID) ? 't' : '-',
		in->numLinks, (uint32_t)in->size, in->numBlocks,
		
		in->directBlocks[0],
		in->directBlocks[0] ? 
			VFS_SECTOR_SIZE * (0x8000 + ufsPrvDataBlockNoToSector(vol, in->directBlocks[0]))
			: 0,
		
		in->directBlocks[1],
		in->directBlocks[1] ? 
			VFS_SECTOR_SIZE * (0x8000 + ufsPrvDataBlockNoToSector(vol, in->directBlocks[1]))
			: 0
		
		);
}

static bool ufsPrvMount(struct VfsVolume *vol)
{
	struct UfsSuper0 *s0 = (struct UfsSuper0*)mBuf;
	struct UfsSuper2 *s2 = (struct UfsSuper2*)mBuf;
	uint32_t i, j;
	
	mBufVol = NULL;	//we're about to read into it, so mark is as invalid
	
	if (sizeof(struct UfsSuper0) != VFS_SECTOR_SIZE) {
		pr("bad size0: %u\n", sizeof(struct UfsSuper0));
		return false;
	}
	
	if (sizeof(struct UfsSuper1) != VFS_SECTOR_SIZE) {
		pr("bad size1: %u\n", sizeof(struct UfsSuper1));
		return false;
	}
	
	if (sizeof(struct UfsSuper2) > VFS_SECTOR_SIZE) {
		pr("bad size2: %u\n", sizeof(struct UfsSuper2));
		return false;
	}
	if (!(s2 = ufsPrvReadSector(vol, SUPERBLOCK_SECTOR + 2)) || s2->magic != UFS_MAGIC || !(s0 = ufsPrvReadSector(vol, SUPERBLOCK_SECTOR + 0)))
		return false;
	
	if (s0->super0ofst >= s0->cylOfst || s0->cylOfst >= s0->inodesOfst || s0->inodesOfst >= s0->firstDataOfst || s0->fsToDb > 8 /* unlikely */)
		return false;

	vol->blksPerGroup_M_fsFrag = s0->blksPerGroup_M_fsFrag;
	vol->cgMask = s0->cgMask;
	vol->cgOffset = s0->cgOffset;
	vol->inodesOfst = s0->inodesOfst;
	vol->firstDataOfst = s0->firstDataOfst;
	vol->numCylGroups = s0->numCylGroups;
	vol->inodesPerGroup = s0->inodesPerGroup;
	vol->basicBlockSz = s0->basicBlockSz;
	vol->fsToDb = s0->fsToDb;
		
	pr("This might be UFS. %u blocks (%u data), %u cyl groups\n",
		s0->numBlocks, s0->numDataBlocks, s0->numCylGroups);
	
	if (0) {
		
		pr("actual offset of root dir is 0x1028000, sec 0x%08x\n", 0x1028000 / VFS_SECTOR_SIZE);
		
		for (i = 0; i < vol->numCylGroups; i++) {
			struct UfsInode *in;
			
			pr("CG %u: start @ 0x%08x, inodes @ 0x%08x, data @0x%08x\n",
				i, ufsPrvBlockToSector(vol, ufsPrvGetCylGrpStart(vol, i)),
				ufsPrvBlockToSector(vol, ufsPrvGetInodesStart(vol, i)), ufsPrvBlockToSector(vol, ufsPrvGetDataStart(vol, i)));
			
			in = ufsPrvReadSector(vol, ufsPrvBlockToSector(vol, ufsPrvGetInodesStart(vol, i)));
			if (!in)
				pr(" cannot read at inodes\n");
			
			for (j = 0; j < 4; j++)
				ufsPrvDumpInode(vol, in + j);
		}
	}
	if (0) {
		const char *interstingDir = "tmp", *interestingFile = "file";
		struct VfsFile *d;
		char name[VFS_SECTOR_SIZE + 1];
		struct VfsFileOpenInfoOpaque oi, oiDir, oiFile;
		uint32_t size, nFiles;
		uint8_t type;
		
		d = vfsOpenRoot(vol);
		pr("open root = 0x%08x\n", d);
		nFiles = 0;
		while (vfsDirEnum(d, name, &oi)) {
			
			if (!vfsFileInfo(vol, &oi, &type, &size))
				break;
			
			pr("found file type %u, size %u inode %u '%s'\r\n", type, size, oi.opaque[0], name);
			
			if (!strcmp(name, interstingDir))
				oiDir = oi;
			if (!strcmp(name, interestingFile))
				oiFile = oi;
			nFiles++;
		}
		pr("enum over with %u entries\n", nFiles);
		vfsFileClose(d);
	
		d = vfsFileOpen(vol, &oiDir);
		pr("open '%s' = 0x%08x\n", interstingDir, d);
		nFiles = 0;
		while (vfsDirEnum(d, name, &oi)) {
			
			if (!vfsFileInfo(vol, &oi, &type, &size))
				break;
			
			pr("found file type %u, size %u inode %u '%s'\r\n", type, size, oi.opaque[0], name);
			nFiles++;
		}
		pr("enum over with %u entries\n", nFiles);
		vfsFileClose(d);
		
		if (!vfsFileInfo(vol, &oiFile, &type, &size) || type != VFS_TYPE_FILE)
			pr("cannot open this file\n");
		else {
	
			uint32_t i;
			
			d = vfsFileOpen(vol, &oiFile);
			pr("open '%s' = 0x%08x, size is %u\n", interestingFile, d, size);
			
			for (i = 0; i < size; i += 4096) {
				
				char ch[5] = {};
				
				if (4 != vfsFileRead(d, &ch, 4)) {
					
					pr("cannot read @ %u\n", i);
					break;
				}
				if (4092 != vfsFileSkip(d, 4092)) {
					
					pr("cannot skip @ %u\n", i);
					break;
				}
				pr("%u -> '%s' (%02x %02x %02x %02x)\n", i, ch, ch[0], ch[1], ch[2], ch[3]);
			}
		
			vfsFileClose(d);
		}
	}

	return true;
}

static struct VfsFile* ufsPrvGetFreeFile(void)
{
	uint_fast8_t i;
	
	for (i = 0; i < MAX_FILES; i++) {
		if (mFilesAvail & (1 << i)) {
			mFilesAvail &=~ (1 << i);
			return mFiles + i;
		}
	}
	
	return NULL;
}

void vfsFileClose(struct VfsFile *file)
{
	mFilesAvail |= 1 << (file - mFiles);
}

static uint32_t ufsPrvFileRead(struct VfsFile *f, void *dstP, uint32_t nBytes)
{
	uint32_t bytesLeftInFile, nBytesOrig = nBytes, blkSz;
	struct VfsVolume *vol = f->vol;
	char *dst = (char*)dstP, *src;
	
	blkSz = vol->basicBlockSz;
	
	while ((bytesLeftInFile = f->size - f->curPos) && nBytes) {
		
		uint_fast16_t ofstInCurBlk = f->curPos % blkSz;
		uint_fast16_t bytesLeftCurSec = VFS_SECTOR_SIZE - ofstInCurBlk % VFS_SECTOR_SIZE;
				
		if (bytesLeftCurSec > bytesLeftInFile)
			bytesLeftCurSec = bytesLeftInFile;
		if (bytesLeftCurSec > nBytes)
			bytesLeftCurSec = nBytes;
		if (dst) {
			
			uint32_t sectorNum = ufsPrvBlockToSector(vol, f->curBlk) + ofstInCurBlk / VFS_SECTOR_SIZE;
			uint32_t ofstInSector = ofstInCurBlk % VFS_SECTOR_SIZE;

			src = ufsPrvReadSector(vol, sectorNum);
			if (!src)
				break;
	
			memcpy(dst, src + ofstInSector, bytesLeftCurSec);
			
			dst += bytesLeftCurSec;
		}
		f->curPos += bytesLeftCurSec;
		ofstInCurBlk += bytesLeftCurSec;
		nBytes -= bytesLeftCurSec;
		
		if (ofstInCurBlk == blkSz) {	//need next block
			
			uint32_t *indir;
			
			if (f->curDirIdx != DIRECT_BLKS_PER_INODE) {
				f->curDirIdx++;
				
				if (f->curDirIdx != DIRECT_BLKS_PER_INODE) {
					
					f->curBlk = f->directBlocks[f->curDirIdx];
					continue;
				}
				//we are about to go indirect!
				f->curIndirIdx = 0;
				f->curIndirIndirIdx = 0;
			}
			else {
				f->curIndirIndirIdx++;
				
				if (f->curIndirIndirIdx * sizeof(uint32_t) == blkSz) {
					pr("going past indir\n");
					f->curIndirIdx++;			//XXX: to what limits?
					f->curIndirIndirIdx = 0;
				}
			}
			
			indir = ufsPrvReadSector(vol, ufsPrvBlockToSector(vol, f->indirectBlocks[f->curIndirIdx]) + f->curIndirIndirIdx * sizeof(uint32_t) / VFS_SECTOR_SIZE);
			indir += (f->curIndirIndirIdx * sizeof(uint32_t) % VFS_SECTOR_SIZE) / sizeof(uint32_t);
			f->curBlk = *indir;
			
		//	pr("indir %u.%u -> [0x%08x] -> 0x%04x\n",
		//		f->curIndirIdx, f->curIndirIndirIdx,
		//		ufsPrvBlockToSector(vol, f->indirectBlocks[f->curIndirIdx]) + f->curIndirIndirIdx * sizeof(uint32_t) / VFS_SECTOR_SIZE,
		//		f->curBlk);
		}
	}
	
	return nBytesOrig - nBytes;
}

static uint32_t ufsPrvFileSkip(struct VfsFile *f, uint32_t numBytesToSkip)
{
	return ufsPrvFileRead(f, NULL, numBytesToSkip);
}

uint32_t vfsFileRead(struct VfsFile *f, void *buf, uint32_t len)
{
	if (f->fileType != VFS_TYPE_FILE)
		return 0;
	
	return ufsPrvFileRead(f, buf, len);
}

uint32_t vfsFileSkip(struct VfsFile *f, uint32_t len)
{
	if (f->fileType != VFS_TYPE_FILE)
		return 0;
	
	return ufsPrvFileSkip(f, len);
}

uint32_t vfsFileTell(const struct VfsFile *f)
{
	if (f->fileType != VFS_TYPE_FILE)
		return 0;
	
	return f->curPos;
}

bool vfsFileSeek(struct VfsFile *f, uint32_t pos)
{
	uint32_t by;
	
	if (f->fileType != VFS_TYPE_FILE)
		return 0;
	
	if (pos >= f->curPos) {
		
		by = pos - f->curPos;
		
		return by == ufsPrvFileSkip(f, by);
	}
	
	//going backwards - be lazy and do not try going back beyond the start of this block
	if ((f->curPos ^ pos) % f->vol->basicBlockSz) {
		
		vfsFileRewind(f);
		return pos == ufsPrvFileSkip(f, pos);
	}
	
	//going back in current block
	f->curPos = pos;
	
	return true;
}

void vfsFileRewind(struct VfsFile *f)
{
	if (f->fileType != VFS_TYPE_FILE)
		return;
	
	f->curPos = 0;
	f->curBlk = f->directBlocks[0];
	f->curDirIdx = 0;
}

static struct UfsInode* ufsPrvGetInode(struct VfsVolume *vol, uint32_t inodeNum)
{
	uint32_t cylGrp, inodeIdx, inodeSec, inodeIdxInSec;
	struct UfsInode *in;
	
	cylGrp = inodeNum / vol->inodesPerGroup;
	inodeIdx = inodeNum % vol->inodesPerGroup;
	inodeSec = inodeIdx / (VFS_SECTOR_SIZE / sizeof(struct UfsInode));
	inodeIdxInSec = inodeIdx % (VFS_SECTOR_SIZE / sizeof(struct UfsInode));
	
	in = ufsPrvReadSector(vol, ufsPrvBlockToSector(vol, ufsPrvGetInodesStart(vol, cylGrp)) + inodeSec);
	
	if (!in)
		return NULL;
	
	return in + inodeIdxInSec;
}

bool vfsFileInfo(struct VfsVolume* vol, struct VfsFileOpenInfoOpaque *oiP, uint8_t *typeP, uint32_t *sizeP)
{
	struct VfsFileOpenInfo *oi = (struct VfsFileOpenInfo*)oiP;
	struct UfsInode *in;
	
	if (typeP || sizeP) {
	
		//we'll need to read the inode to know more...
		in = ufsPrvGetInode(vol, oi->inode);
		if (!in)
			return false;
		
//		pr("file '%s' at inode 0x%08x\n", nameP ? nameP : "<not read>", de.inode);
		

//		pr("inode info for that file:\n");
//		ufsPrvDumpInode(vol, in);
		
		if (typeP)
			*typeP = in->type;
		if (sizeP)
			*sizeP = in->size;
	}
	
	return true;
}

bool vfsDirEnum(struct VfsFile *dir, char *nameP, struct VfsFileOpenInfoOpaque *oiP)
{
	struct VfsFileOpenInfo *oi = (struct VfsFileOpenInfo*)oiP;
	#if defined(UFS_TYPE_OLD)
		struct UfsDirEntry de;
	#elif defined(UFS_TYPE_FFS)
		struct FfsDirEntry de;
	#endif
	uint_fast16_t toSkip;
	
	if (!dir->curBlk)
		return false;
	
	if (!ufsPrvFileRead(dir, &de, sizeof(de)))
		return false;
	
	if (de.nameLen > VFS_MAX_NAME_LEN) {
		pr("invalid name length 0x%04x\n", de.nameLen);
		return false;	//and we'll crash soon after!
	}
	
	toSkip = de.structSz - sizeof(struct UfsDirEntry);
	if (nameP) {
		
		if (!ufsPrvFileRead(dir, nameP, de.nameLen + 1))
			return false;
		toSkip -= de.nameLen + 1;
	}
	
	if (toSkip != ufsPrvFileSkip(dir, toSkip))
		return false;	//and we'll crash soon after!
	
	if (oi)
		oi->inode = de.inode;
	
	return true;
}

static struct VfsFile* ufsPrvOpenFileWithByInode(struct VfsVolume *vol, uint32_t inodeNum, uint_fast16_t allowedType /* bitmask of 1 << VFS_TYPE_* */)
{
	struct UfsInode* in = ufsPrvGetInode(vol, inodeNum);
	struct VfsFile *f;
	uint_fast8_t i;
		
	if (!in)
		return NULL;
	
	if (!(allowedType & (1 << in->type)) || in->size >> 32)
		return NULL;
	
	f = ufsPrvGetFreeFile();
	if (!f)
		return NULL;
		
	f->vol = vol;
	f->size = in->size;
	for (i = 0; i < DIRECT_BLKS_PER_INODE; i++) {
		f->directBlocks[i] = in->directBlocks[i];
//		pr(" direct block[%u] = 0x%08x\n", i, f->directBlocks[i]);
	}
	for (i = 0; i < INDIR_BLKS_PER_INODE; i++)
		f->indirectBlocks[i] = in->indirectBlocks[i];
	
	f->curPos = 0;
	f->curBlk = f->directBlocks[0];
	f->curDirIdx = 0;
	f->fileType = in->type;
	
	return f;
}

struct VfsFile* vfsFileOpen(struct VfsVolume* vol, struct VfsFileOpenInfoOpaque *oiP)
{
	const struct VfsFileOpenInfo *oi = (const struct VfsFileOpenInfo*)oiP;
	
	return ufsPrvOpenFileWithByInode(vol, oi->inode, (1 << VFS_TYPE_DIR) | (1 << VFS_TYPE_FILE));
}

struct VfsFile* vfsOpenRoot(struct VfsVolume *vol)
{
	return ufsPrvOpenFileWithByInode(vol, 2, 1 << VFS_TYPE_DIR);
}

struct VfsVolume* vfsMount(VfsReadSecProc readSecF, void *readSecD)
{
	uint_fast8_t i;
	
	for (i = 0; i < 8; i++) {
		
		if (mVolumesAvail & (1 << i)) {
			
			mVolumes[i].readF = readSecF;
			mVolumes[i].readD = readSecD;
			
			if (ufsPrvMount(mVolumes + i)) {
				
				mVolumesAvail &=~ (1 << i);
				return mVolumes + i;
			}
			
			break;
		}
	}
	
	return NULL;
}

void vfsInit(void *tmpBuf)
{
	mBuf = tmpBuf;
}

void vfsUnmount(struct VfsVolume *vol)
{
	mVolumesAvail |= 1 << (vol - mVolumes);
}